(*------------------------------------------------------------------------

 TBrowseDialog Compenent for Delphi 2.0
 Versin 1.4

 Written by Shugo Maeda
 NOT ANY Rights Reserved.
 ~~~~~~~
 Work with each other rather than against!


 History

   10/09/96 v1.0
   10/11/96 v1.1
   10/24/96 v1.2
   11/26/96 v1.3
   12/05/96 v1.4

 ------------------------------------------------------------------------*)

unit BrowsDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  SHLOBJ, OLE2;

type
  TBrowseOption = (biBrowseForComputer, biBrowseForPrinter,
    biDontGoBelowDomain, biReturnFSAncestors, biReturnOnlyFSDirs,
    biStatusText);
  TBrowseOptions = set of TBrowseOption;

  TOKButtonState = (obsDefault, obsEnabled, obsDisabled);

  TBrowseDialogEvent = procedure (Sender: TObject; Wnd: HWND;
    Path: string;var ShowText: string;var OKButtonState: TOKButtonState) of object;

  TBrowseDialog = class(TComponent)
  private
    FDisplayName: string;
    FDirectory: string;
    FCaption: string;
    FOptions: TBrowseOptions;
    FRoot: string;
    FOnInitialized: TBrowseDialogEvent;
    FOnSelectionChanged: TBrowseDialogEvent;
  protected
    function Message(var Msg: TMessage):Boolean; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    property DisplayName: string read FDisplayName;
    function Execute: Boolean; virtual;
    procedure SelectionChanged(Wnd: HWND; Path: string;var ShowText: string;
      var OKButtonState: TOKButtonState);
    procedure Initialized(Wnd: HWND; Path: string;var ShowText: string;
      var OKButtonState: TOKButtonState);
  published
    property Directory: string read FDirectory write FDirectory;
    property Caption: string read FCaption write FCaption;
    property Root: string read FRoot write FRoot;
    property Options: TBrowseOptions read FOptions write FOptions default [biReturnOnlyFSDirs];
    property OnInitialized: TBrowseDialogEvent read FOnInitialized write FOnInitialized;
    property OnSelectionChanged: TBrowseDialogEvent read FOnSelectionChanged write FOnSelectionChanged;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Dialogs', [TBrowseDialog]);
end;

constructor TBrowseDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOptions := [biReturnOnlyFSDirs];
end;

procedure TBrowseDialog.Initialized(Wnd: HWND; Path: string;
            var ShowText: string;var OKButtonState: TOKButtonState);
begin
  if Assigned(FOnInitialized) then
    FOnInitialized(Self, Wnd, Path, ShowText, OKButtonState);
end;

procedure TBrowseDialog.SelectionChanged(Wnd: HWND; Path: string;
            var ShowText: string;var OKButtonState: TOKButtonState);
begin
  if Assigned(FOnSelectionChanged) then
    FOnSelectionChanged(Self, Wnd, Path, ShowText, OKButtonState);
end;

// Callback function called from SHBrowseForFolder()
function BrowseCallbackProc(Wnd: HWND; uMsg: UINT;
             lParam, lpData: LPARAM): Integer stdcall;
var
  Path: array[0..512] of Char;
  ShowText: string;
  OKButtonState: TOKButtonState;
  Dir: string;
begin
  Result := 0;
  if not (TObject(lpData) is TBrowseDialog) then Exit;
  Path := '';
  ShowText := '';
  OKButtonState := obsDefault;
  if uMsg = BFFM_SELCHANGED then
  begin
    SHGetPathFromIDList(PItemIDList(lParam), Path);
    TBrowseDialog(lpData).SelectionChanged(Wnd,
      string(Path), ShowText, OKButtonState);
  end;
  if uMSG = BFFM_INITIALIZED then
  begin
    if TBrowseDialog(lpData).Directory <> '' then
    begin
      Dir := TBrowseDialog(lpData).Directory;
      if (Length(Dir) > 3) and (Dir[Length(Dir)] = '\') then
        Delete(Dir, Length(Dir), 1);
      SendMessage(Wnd, BFFM_SETSELECTION, Longint(True), Longint(PChar(Dir)));
    end;
    TBrowseDialog(lpData).Initialized(Wnd, Dir, ShowText, OKButtonState);
  end;
  SendMessage(Wnd, BFFM_SETSTATUSTEXT, 0, Longint(PChar(ShowText)));
  case OKButtonState of
    obsEnabled:
      SendMessage(Wnd, BFFM_ENABLEOK, 0, Longint(True));
    obsDisabled:
      SendMessage(Wnd, BFFM_ENABLEOK, 0, Longint(False));
  end;
end;

function TBrowseDialog.Message(var Msg: TMessage): Boolean;
begin
  Result := False;
end;

// Execute Method
function TBrowseDialog.Execute: Boolean;
const
  BrowseOptions: array[TBrowseOption] of UINT = (
    BIF_BROWSEFORCOMPUTER, BIF_BROWSEFORPRINTER ,
    BIF_DONTGOBELOWDOMAIN, BIF_RETURNFSANCESTORS,
    BIF_RETURNONLYFSDIRS, BIF_STATUSTEXT);
var
  ppmem: IMalloc;
  lpbuffer: LPSTR;
  BrowseInfo: TBrowseInfo;
  pidl, pidlBrowse: PItemIDList;
  Res: HResult;
  ppshf: IShellFolder;
  lpszDisplayName: POLESTR;
  pchEaten, dwAttributes: ULONG;
  Option: TBrowseOption;
  ActiveWindow: HWnd;
  WindowList: Pointer;
begin
  Result := False;

  if SHGetMalloc(ppmem) = S_OK then
  begin
    try
      // Get PIDL for the root directory.
      pidl := nil; // if "Root" property is empty, Desktop is root.
      if FRoot <> '' then
      begin
        Res:= SHGetDesktopFolder(ppshf);
        if Failed(Res) then
          raise Exception.Create('TBrowseDialog: SHGetDesktopFolder Failed.');
        try
          lpszDisplayName := ppmem.Alloc(MAX_PATH);
          if lpszDisplayName = nil Then
            raise Exception.Create('TBrowseDialog: ppmem.Alloc Failed.');

          lpszDisplayName:= StringToOleStr(FRoot);
          pchEaten:= SizeOf(lpszDisplayName) div 2;

          Res := ppshf.ParseDisplayName(Application.Handle,
                                       nil,
                                       lpszDisplayName,
                                       pchEaten,
                                       pidl,
                                       dwAttributes);
          if Failed(Res) then pidl := nil;
          ppmem.Free(lpszDisplayName);
        finally
          ppshf.Release();
        end;
      end;

      lpbuffer := ppmem.Alloc(MAX_PATH);
      if lpbuffer = nil then
        raise Exception.Create('TBrowseDialog: ppmem.Alloc Failed.');

      try
        // Set BrowseInfo
        with BrowseInfo do
        begin
          hwndOwner := Application.Handle;
          pidlRoot := pidl;
          pszDisplayName := lpbuffer;
          lpszTitle  := PChar(FCaption);
          ulFlags := 0;
          for Option := Low(Option) to High(Option) do
            if Option in FOptions then
              ulFlags := ulFlags or BrowseOptions[Option];
          lpfn := BrowseCallbackProc;
          lParam := Longint(Self);
        end;

        ActiveWindow := GetActiveWindow;
        WindowList := DisableTaskWindows(0);
        try
          Application.HookMainWindow(Message);
          try
            pidlBrowse := SHBrowseForFolder(BrowseInfo);
          finally
            Application.UnhookMainWindow(Message);
          end;
        finally
          EnableTaskWindows(WindowList);
          SetActiveWindow(ActiveWindow);
        end;
        if pidlBrowse <> nil then
        begin
          Result := True;
          FDisplayName := string(lpbuffer);
          if SHGetPathFromIDList(pidlBrowse, lpbuffer) = True then
            FDirectory := string(lpbuffer)
          else
            FDirectory := '';
          ppmem.Free(pidlBrowse);
        end;
      finally
        ppmem.Free(pidl);
        ppmem.Free(lpbuffer);
      end;
    finally
      ppmem.Release();
    end;
  end;
end;

end.

